; #########################################################################
;
;                             ZTanh
;
;##########################################################################

  ; -----------------------------------------------------------------------
  ; This procedure was written by Raymond Filiatreault, February 2005
  ;
  ;                     tan(Src) -> Dest
  ;  starting with cartesian coordinates
  ;   = sinh(2a)/[cosh(2a)+cos(2b)] , sin(2b)/[cosh(2a)+cos(2b)]
  ;              Note: (much faster than sinh/cosh)
  ;
  ; This ZTanh function computes the hyperbolic tangent of the complex
  ; number Src and returns the result as a complex number at the specified
  ; destination (the FPU itself or a memory location), unless an invalid
  ; operation is detected or the definition of the parameters (with uID)
  ; is invalid.
  ;
  ; The source can be either in cartesian or polar coordinates. The result
  ; will be returned in the same coordinates.
  ;
  ; The source can be a complex number already on the FPU or from memory. If
  ; on the FPU, the REAL coefficient must be in the ST(0) register and the
  ; IMAGINARY coefficient in the ST(1) register. Memory variables must have
  ; the two coefficients at adjacent memory addresses, the REAL coefficient
  ; being at the lower address. All memory sources and destinations must
  ; have the same declared size as specified in the uID parameter.
  ;
  ; The source is not checked for validity. This is the programmer's
  ; responsibility. The programmer is also responsible to insure that all
  ; the FPU data registers necessary for performing the function are FREE.
  ;   If either Src1 or Src2 is on the FPU:  ST(5), ST(6) and ST(7)
  ;    must be free
  ;   Otherwise: ST(3), ST(4), ST(5), ST(6) and ST(7) must be free
  ;
  ; Only EAX is used to return error (0) or success (!0). All other CPU
  ; registers are preserved.
  ;
  ; If a valid operand size is present in the uID, any source data on the
  ; FPU will be modified; otherwise, if an invalid operand size is present
  ; in the uID, all FPU registers are preserved and the location of the
  ; destination is not altered.
  ;
  ; If an invalid operation is detected, the modified data would be
  ; considered useless and the top two registers would be popped; also, if
  ; a memory location is specified as the destination, it would not be
  ; modified.
  ;
  ; -----------------------------------------------------------------------

    .386
    .model flat, stdcall  ; 32 bit memory model
    option casemap :none  ; case sensitive

    include ZLIB.inc

    .code

;#######################################################################

ZTanh proc public USES ecx lpSrc1:DWORD, lpDest:DWORD, uID:DWORD
        
;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

      mov   ecx,uID
      and   ecx,Z_SINGLE or Z_DBL   ;check uID for operand size
      .if   ZERO?
            mov   ecx,10
            jmp   @F
      .elseif ecx == 4
            jmp   @F
      .elseif ecx == 8
            jmp   @F
      .endif
      xor   eax,eax                 ;error code
      ret                           ;no valid operand size

   @@:
      test  uID,SRC1_FPU            ;check if Src1 in on FPU
      .if   ZERO?                   ;Src1 not on FPU
            mov   eax,lpSrc1
            .if   ecx == 4
                  fld   dword ptr [eax+ecx]     ;imag1
                  fld   dword ptr [eax]         ;real1   imag1
            .elseif ecx == 8
                  fld   qword ptr [eax+ecx]
                  fld   qword ptr [eax]
            .elseif ecx == 10
                  fld   tbyte ptr [eax+ecx]
                  fld   tbyte ptr [eax]
            .endif
      .endif

; The exponential and addition functions will need cartesian coordinates

      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            invoke ZPtoC,0,0,SRC1_FPU or DEST_FPU     ;->cartesian
      .endif

                        ;a    b
      fxch              ;b    a
      fadd  st,st       ;2b   b
      fsincos           ;cos(2b)   sin(2b)    a
      fxch  st(2)       ;a    sin(2b)    cos(2b)

;=======
; e^(2a)
;=======
      fadd  st,st       ;2a    sin(2b)    cos(2b)
      fldl2e            ;log2(e)   2a    sin(2b)    cos(2b)
      fmul              ;log2(e)*2a    sin(2b)    cos(2b)
      fld   st          ;log2(e)*2a    log2(e)*2a    sin(2b)    cos(2b)
      frndint           ;int[log2(e)*2a]    log2(e)*2a    sin(2b)    cos(2b)
      fxch              ;log2(e)*2a    int[log2(e)*2a]    sin(2b)    cos(2b)
      fsub  st,st(1)    ;mod   int[log2(e)*2a]    sin(2b)    cos(2b)
      f2xm1             ;2^mod-1   int[log2(e)*2a]    sin(2b)    cos(2b)
      fld1              ;1    2^mod-1   int[log2(e)*2a]    sin(2b)    cos(2b)
      fadd              ;2^mod   int[log2(e)*2a]    sin(2b)    cos(2b)
      fscale            ;e^2a   int[log2(e)*2a]    sin(2b)    cos(2b)
      fstp  st(1)       ;e^2a    sin(2b)    cos(2b)

;==================
; sinh(2a) cosh(2a)
;==================
      fld1              ;1     e^2a    sin(2b)    cos(2b)
      fdiv  st,st(1)    ;e^-2a  e^2a    sin(2b)    cos(2b)
      fld   st(1)       ;e^2a   e^-2a  e^2a    sin(2b)    cos(2b)
      fadd  st,st(1)    ;e^2a+e^-2a  e^-2a  e^2a    sin(2b)    cos(2b)
      fxch  st(2)       ;e^2a  e^-2a  e^2a+e^-2a    sin(2b)    cos(2b)
      fsubr             ;e^2a-e^-2a  e^2a+e^-2a    sin(2b)    cos(2b)
      fld1              ;1    e^a-e^-2a  e^a+e^-2a    sin(2b)    cos(2b)
      fchs              ;-1   e^a-e^-2a  e^a+e^-2a    sin(2b)    cos(2b)
      fxch              ;e^2a-e^-2a   -1  e^2a+e^-2a    sin(2b)    cos(2b)
      fscale            ;sinh(2a)   -1  e^a+e^-2a    sin(2b)    cos(2b)
      fxch  st(2)       ;e^2a+e^-2a   -1   sinh(2a)   sin(2b)    cos(2b)
      fscale            ;cosh(2a)   -1   sinh(2a)   sin(2b)    cos(2b)
      fstp  st(1)       ;cosh(2a)   sinh(2a)   sin(2b)    cos(2b)

;=========================================================
; sinh(2a)/[cosh(2a)+cos(2b)] , sin(2b)/[cosh(2a)+cos(2b)]
;=========================================================
      faddp st(3),st    ;sinh(2a)   sin(2b)    [cosh(2a)+cos(2b)]
      fdiv  st,st(2)    ;sinh(2a)/[cosh(2a)+cos(2b)]   sin(2b)    [cosh(2a)+cos(2b)]
      fxch  st(2)       ;[cosh(2a)+cos(2b)]    sin(2b)   sinh(2a)/[cosh(2a)+cos(2b)]
      fdiv              ;sin(2b)/[cosh(2a)+cos(2b)]   sinh(2a)/[cosh(2a)+cos(2b)]
      fxch              ;a    b

      test  uID,Z_POLAR             ;polar coordinates?
      .if   !ZERO?
            invoke ZCtoP,0,0,SRC1_FPU or DEST_FPU     ;->polar
      .endif

      fstsw ax                      ;get exception flags
      fwait
      shr   al,1                    ;test for invalid operation
      .if   CARRY?                  ;indicates invalid operation
            fstp  st                ;pop top register
            fstp  st                ;pop second register
            xor   eax,eax           ;error return code
            ret
      .endif

      test  uID,DEST_FPU            ;check destination of result
      .if   ZERO?                   ;to be stored in memory
            mov   eax,lpDest
            .if   ecx == 4
                  fstp  dword ptr [eax]         ;store REAL first
                  fstp  dword ptr [eax+ecx]     ;then store IMAG
            .elseif ecx == 8
                  fstp  qword ptr [eax]
                  fstp  qword ptr [eax+ecx]
            .else
                  fstp  tbyte ptr [eax]
                  fstp  tbyte ptr [eax+ecx]
            .endif
      .endif

      or    al,1                    ;to insure EAX!=0
      ret
    
ZTanh endp

; #########################################################################

end
